@@ -1,6 +1,6 @@
 ;;; exwm-core.el --- Core definitions  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2016 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -38,6 +38,9 @@
   (when exwm-debug-on
     `(message (concat "[EXWM] " ,format-string) ,@args)))
 
+(defmacro exwm--debug (&rest forms)
+  (when exwm-debug-on `(progn ,@forms)))
+
 (defvar exwm--connection nil "X connection.")
 (defvar exwm--root nil "Root window.")
 (defvar exwm--id-buffer-alist nil "Alist of (<X window ID> . <Emacs buffer>).")
@@ -75,15 +78,18 @@
     (logior xcb:EventMask:StructureNotify xcb:EventMask:PropertyChange))
   "Event mask set on all managed windows.")
 
+(declare-function exwm-input--on-KeyPress-line-mode "exwm-input.el"
+                  (key-press))
+
 ;; Internal variables
-(defvar-local exwm--id nil)                        ;window ID
-(defvar-local exwm--frame nil)                     ;workspace frame
-(defvar-local exwm--floating-frame nil)            ;floating frame
-(defvar-local exwm--floating-edges nil)            ;four edges
-(defvar-local exwm--floating-mode-line-format nil) ;save mode-line-format
-(defvar-local exwm--fullscreen nil)                ;used in fullscreen
-(defvar-local exwm--floating-frame-geometry nil)   ;in fullscreen
-(defvar-local exwm--fixed-size nil)                ;fixed size
+(defvar-local exwm--id nil)               ;window ID
+(defvar-local exwm--container nil)        ;container
+(defvar-local exwm--frame nil)            ;workspace frame
+(defvar-local exwm--floating-frame nil)   ;floating frame
+(defvar-local exwm--mode-line-format nil) ;save mode-line-format
+(defvar-local exwm--fullscreen nil)       ;used in fullscreen
+(defvar-local exwm--floating-frame-position nil) ;used in fullscreen
+(defvar-local exwm--fixed-size nil)              ;fixed size
 (defvar-local exwm--on-KeyPress         ;KeyPress event handler
   #'exwm-input--on-KeyPress-line-mode)
 ;; Properties
@@ -107,21 +113,24 @@
 (defvar-local exwm--normal-hints-max-height nil)
 ;; (defvar-local exwm--normal-hints-win-gravity nil)
 ;; WM_HINTS
-(defvar-local exwm--hints-input nil)    ;FIXME
+(defvar-local exwm--hints-input nil)
 (defvar-local exwm--hints-urgency nil)
 ;; _MOTIF_WM_HINTS
 (defvar-local exwm--mwm-hints nil)
 
 (defvar exwm-mode-map
   (let ((map (make-sparse-keymap)))
-    (define-key map "\C-ck" #'exwm-input-release-keyboard)
-    (define-key map "\C-cf" #'exwm-layout-set-fullscreen)
-    (define-key map "\C-cm" #'exwm-floating-toggle-floating)
-    (define-key map "\C-cq" #'exwm-input-send-next-key)
-    (define-key map "\C-cv" #'exwm-workspace-move-window)
+    (define-key map "\C-c\C-f" #'exwm-layout-set-fullscreen)
+    (define-key map "\C-c\C-k" #'exwm-input-release-keyboard)
+    (define-key map "\C-c\C-m" #'exwm-workspace-move-window)
+    (define-key map "\C-c\C-q" #'exwm-input-send-next-key)
+    (define-key map "\C-c\C-t\C-f" #'exwm-floating-toggle-floating)
+    (define-key map "\C-c\C-t\C-m" #'exwm-layout-toggle-mode-line)
     map)
   "Keymap for `exwm-mode'.")
 
+(declare-function exwm-manage--kill-buffer-query-function "exwm-manage.el")
+
 (define-derived-mode exwm-mode nil "EXWM"
   "Major mode for managing X windows.
 
@@ -138,9 +147,7 @@
   (add-hook 'change-major-mode-hook #'kill-buffer nil t)
   ;; Kill buffer -> close window
   (add-hook 'kill-buffer-query-functions
-            (lambda ()
-              (exwm-manage--close-window exwm--id (current-buffer)))
-            nil t)
+            #'exwm-manage--kill-buffer-query-function nil t)
   (setq buffer-read-only t
         left-margin-width nil
         right-margin-width nil
