@@ -1,6 +1,6 @@
 ;;; exwm-randr.el --- RandR Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2016 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -48,13 +48,22 @@
 
 (require 'xcb-randr)
 (require 'exwm-core)
-(eval-when-compile (require 'exwm-workspace))
 
 (defvar exwm-randr-workspace-output-plist nil)
 
+(defvar exwm-randr-refresh-hook nil
+  "Normal hook run when the RandR module just refreshed.")
+
+(defvar exwm-layout--fullscreen-frame-count)
+(defvar exwm-workspace-number)
+(defvar exwm-workspace--list)
+
+(declare-function exwm-layout--set-frame-fullscreen "exwm-layout.el" (frame))
+
 (defun exwm-randr--refresh ()
   "Refresh workspaces according to the updated RandR info."
-  (let (output-name geometry output-plist default-geometry workareas viewports)
+  (let (output-name geometry output-plist default-geometry workareas
+                    workarea-offset viewports)
     ;; Query all outputs
     (with-slots (config-timestamp outputs)
         (xcb:+request-unchecked+reply exwm--connection
@@ -82,38 +91,38 @@
                     output-plist (plist-put output-plist output-name geometry))
               (unless default-geometry ;assume the first output as primary
                 (setq default-geometry geometry)))))))
-    (cl-assert (<= 2 (length output-plist)))
-    (dotimes (i exwm-workspace-number)
-      (let* ((output (plist-get exwm-randr-workspace-output-plist i))
-             (geometry (lax-plist-get output-plist output))
-             (frame (elt exwm-workspace--list i)))
-        (unless geometry
-          (setq geometry default-geometry
-                output nil))
-        (set-frame-parameter frame 'exwm-randr-output output)
-        (set-frame-parameter frame 'exwm-geometry geometry)
-        (with-slots (x y width height) geometry
-          (xcb:+request exwm--connection
-              (make-instance 'xcb:ConfigureWindow
-                             :window (frame-parameter frame 'exwm-outer-id)
-                             :value-mask (eval-when-compile
-                                           (logior xcb:ConfigWindow:X
-                                                   xcb:ConfigWindow:Y
-                                                   xcb:ConfigWindow:Width
-                                                   xcb:ConfigWindow:Height))
-                             :x x :y y :width width :height height))
-          (setq workareas (nconc workareas (list x y width height))
-                viewports (nconc viewports (list x y))))))
-    ;; Update _NET_WORKAREA
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:ewmh:set-_NET_WORKAREA
-                       :window exwm--root :data (vconcat workareas)))
-    ;; Update _NET_DESKTOP_VIEWPORT
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:ewmh:set-_NET_DESKTOP_VIEWPORT
-                       :window exwm--root
-                       :data (vconcat viewports)))
-    (xcb:flush exwm--connection)))
+    (exwm--log "(randr) outputs: %s" output-plist)
+    (when output-plist
+      (setq workarea-offset (if (exwm-workspace--minibuffer-own-frame-p)
+                                0
+                              (window-pixel-height (minibuffer-window))))
+      (setq exwm-layout--fullscreen-frame-count 0)
+      (dotimes (i exwm-workspace-number)
+        (let* ((output (plist-get exwm-randr-workspace-output-plist i))
+               (geometry (lax-plist-get output-plist output))
+               (frame (elt exwm-workspace--list i)))
+          (unless geometry
+            (setq geometry default-geometry
+                  output nil))
+          (set-frame-parameter frame 'exwm-randr-output output)
+          (set-frame-parameter frame 'exwm-geometry geometry)
+          (exwm-layout--set-frame-fullscreen frame)
+          (with-slots (x y width height) geometry
+            (setq workareas
+                  (nconc workareas (list x y width (- height
+                                                      workarea-offset)))
+                  viewports (nconc viewports (list x y))))))
+      ;; Update _NET_WORKAREA
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:ewmh:set-_NET_WORKAREA
+                         :window exwm--root :data (vconcat workareas)))
+      ;; Update _NET_DESKTOP_VIEWPORT
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:ewmh:set-_NET_DESKTOP_VIEWPORT
+                         :window exwm--root
+                         :data (vconcat viewports)))
+      (xcb:flush exwm--connection)
+      (run-hooks 'exwm-randr-refresh-hook))))
 
 (defvar exwm-randr-screen-change-hook nil
   "Normal hook run when screen changes.")
@@ -130,6 +139,8 @@
       (if (or (/= major-version 1) (< minor-version 2))
           (error "[EXWM] The server only support RandR version up to %d.%d"
                  major-version minor-version)
+        ;; External monitor(s) may already be connected.
+        (run-hooks 'exwm-randr-screen-change-hook)
         (exwm-randr--refresh)
         (xcb:+event exwm--connection 'xcb:randr:ScreenChangeNotify
                     (lambda (_data _synthetic)
